# Copyright 2023 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import itertools

from alembic import op
from oslo_db import exception as os_db_exc
import sqlalchemy as sa
from sqlalchemy import exc

from neutron._i18n import _


# Drop unused tables
#
# Revision ID: 054e34dbe6b4
# Revises: 89c58a70ceba
# Create Date: 2023-10-05 16:02:03.812939

# revision identifiers, used by Alembic.
revision = '054e34dbe6b4'
down_revision = '89c58a70ceba'

# Ordered tables to delete first the tables with foreign keys references.
REPO_CISCO_TABLES = [
    'cisco_ml2_apic_contracts',
    'cisco_ml2_apic_names',
    'cisco_ml2_apic_host_links',
    'cisco_ml2_n1kv_policy_profiles',
    'cisco_ml2_n1kv_port_bindings',
    'cisco_ml2_n1kv_network_bindings',
    'cisco_ml2_n1kv_vxlan_allocations',
    'cisco_ml2_n1kv_vlan_allocations',
    'cisco_ml2_n1kv_profile_bindings',
    'cisco_ml2_nexusport_bindings',
    'cisco_ml2_nexus_nve',
    'ml2_nexus_vxlan_mcast_groups',
    'ml2_ucsm_port_profiles',
    'cisco_port_mappings',
    'cisco_router_mappings',
    'cisco_hosting_devices',
    'cisco_ml2_n1kv_network_profiles',
    'ml2_nexus_vxlan_allocations',
    'cisco_network_profiles',
    'cisco_policy_profiles',
]

REPO_VMWARE_TABLES = [
    'tz_network_bindings',
    'neutron_nsx_network_mappings',
    'neutron_nsx_security_group_mappings',
    'neutron_nsx_port_mappings',
    'neutron_nsx_router_mappings',
    'multi_provider_networks',
    'networkconnections',
    'networkgatewaydevicereferences',
    'networkgatewaydevices',
    'networkgateways',
    'maclearningstates',
    'portqueuemappings',
    'networkqueuemappings',
    'qosqueues',
    'lsn_port',
    'lsn',
    'nsxv_router_bindings',
    'nsxv_edge_vnic_bindings',
    'nsxv_edge_dhcp_static_bindings',
    'nsxv_internal_networks',
    'nsxv_internal_edges',
    'nsxv_security_group_section_mappings',
    'nsxv_rule_mappings',
    'nsxv_port_vnic_mappings',
    'nsxv_router_ext_attributes',
    'nsxv_tz_network_bindings',
    'nsxv_port_index_mappings',
    'nsxv_firewall_rule_bindings',
    'nsxv_spoofguard_policy_network_mappings',
    'nsxv_vdr_dhcp_bindings',
    'vcns_router_bindings',
]

REPO_BROCADE_TABLES = [
    'brocadeports',
    'brocadenetworks',
    'ml2_brocadeports',
    'ml2_brocadenetworks',
]

REPO_NUAGE_TABLES = [
    'nuage_net_partition_router_mapping',
    'nuage_provider_net_bindings',
    'nuage_subnet_l2dom_mapping',
    'nuage_net_partitions',
]

TABLES_TO_DROP = itertools.chain(
    REPO_CISCO_TABLES,
    REPO_VMWARE_TABLES,
    REPO_BROCADE_TABLES,
    REPO_NUAGE_TABLES,
    ['cisco_csr_identifier_map']
)


def upgrade():
    inspector = sa.inspect(op.get_bind())
    db_tables = inspector.get_table_names()
    tables_to_drop = list(TABLES_TO_DROP)
    while tables_to_drop:
        # Tables that have not been dropped in this iteration.
        missed_tables = []
        errors = []
        for table in (t for t in tables_to_drop if t in db_tables):
            try:
                op.drop_table(table)
            except (exc.OperationalError, os_db_exc.DBError) as _exc:
                # If the table cannot be deleted is because some reference
                # from other table that should be deleted first.
                missed_tables.append(table)
                errors.append(str(_exc))

        if len(missed_tables) == len(tables_to_drop):
            raise Exception(_('The following tables have not been dropped '
                              'from the Neutron database: %(missed_tables)s.\n'
                              'List of errors: %(errors)s' %
                              {'missed_tables': ', '.join(missed_tables),
                               'errors': '\n'.join(errors)}
                              ))

        tables_to_drop = missed_tables


def expand_drop_exceptions():
    """Support dropping TABLES_TO_DROP"""

    return {
        sa.Table: list(TABLES_TO_DROP)
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib import exceptions
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent.common import ovs_lib
from neutron.services.trunk.drivers.openvswitch.agent import exceptions as exc
from neutron.services.trunk.drivers.openvswitch import utils

LOG = logging.getLogger(__name__)


class TrunkManagerError(exceptions.NeutronException):
    message = _("Error while communicating with OVSDB: %(error)s")


def get_br_int_port_name(prefix, port_id):
    """Return the OVS port name for the given port ID.

    The port name is the one that plumbs into the integration bridge.
    """
    return ("%si-%s" % (prefix, port_id))[:constants.DEVICE_NAME_MAX_LEN]


def get_br_trunk_port_name(prefix, port_id):
    """Return the OVS port name for the given port ID.

    The port name is the one that plumbs into the trunk bridge.
    """
    return ("%st-%s" % (prefix, port_id))[:constants.DEVICE_NAME_MAX_LEN]


def get_patch_peer_attrs(peer_name, port_mac=None, port_id=None):
    external_ids = {}
    if port_mac:
        external_ids['attached-mac'] = port_mac
    if port_id:
        external_ids['iface-id'] = port_id
    attrs = [('type', 'patch'),
             ('options', {'peer': peer_name})]
    if external_ids:
        attrs.append(
            ('external_ids', external_ids))
    return attrs


class TrunkBridge(ovs_lib.OVSBridge):
    """An OVS trunk bridge.

    A trunk bridge has a name that follows a specific naming convention.
    """
    def __init__(self, trunk_id):
        name = utils.gen_trunk_br_name(trunk_id)
        super(TrunkBridge, self).__init__(name)

    def exists(self):
        return self.bridge_exists(self.br_name)


class TrunkParentPort(object):
    """An OVS trunk parent port.

    A trunk parent port is represented in OVS with two patch ports that
    connect a trunk bridge and the integration bridge respectively.
    These patch ports follow strict naming conventions: tpi-<hash> for
    the patch port that goes into the integration bridge, and tpt-<hash>
    for the patch port that goes into the trunk bridge.
    """
    DEV_PREFIX = 'tp'

    def __init__(self, trunk_id, port_id, port_mac=None):
        self.trunk_id = trunk_id
        self.port_id = port_id
        self.port_mac = port_mac
        self.bridge = TrunkBridge(self.trunk_id)
        self.patch_port_int_name = get_br_int_port_name(
            self.DEV_PREFIX, port_id)
        self.patch_port_trunk_name = get_br_trunk_port_name(
            self.DEV_PREFIX, port_id)
        self._transaction = None

    def plug(self, br_int):
        """Plug patch ports between trunk bridge and given bridge.

        The method plugs one patch port on the given bridge side using
        port MAC and ID as external IDs.  The other endpoint of patch port is
        attached to the trunk bridge.  Everything is done in a single
        OVSDB transaction so either all operations succeed or fail.

        :param br_int: an integration bridge where peer endpoint of patch port
                       will be created.
        """
        # NOTE(jlibosva): OVSDB is an api so it doesn't matter whether we
        # use self.bridge or br_int
        ovsdb = self.bridge.ovsdb
        # Once the bridges are connected with the following patch ports,
        # the ovs agent will recognize the ports for processing and it will
        # take over the wiring process and everything that entails.
        # REVISIT(rossella_s): revisit this integration part, should tighter
        # control over the wiring logic for trunk ports be required.
        patch_int_attrs = get_patch_peer_attrs(
            self.patch_port_trunk_name, self.port_mac, self.port_id)
        patch_trunk_attrs = get_patch_peer_attrs(self.patch_port_int_name,
                                                 self.port_mac, self.port_id)

        with ovsdb.transaction() as txn:
            txn.add(ovsdb.add_port(br_int.br_name,
                                   self.patch_port_int_name))
            txn.add(ovsdb.db_set('Interface', self.patch_port_int_name,
                                 *patch_int_attrs))
            txn.add(ovsdb.add_port(self.bridge.br_name,
                                   self.patch_port_trunk_name))
            txn.add(ovsdb.db_set('Interface', self.patch_port_trunk_name,
                                 *patch_trunk_attrs))

    def unplug(self, bridge):
        """Unplug the trunk from bridge.

        Method unplugs in single OVSDB transaction the trunk bridge and patch
        port on provided bridge.

        :param bridge: bridge that has peer side of patch port for this
                       subport.
        """
        ovsdb = self.bridge.ovsdb
        with ovsdb.transaction() as txn:
            txn.add(ovsdb.del_port(self.patch_port_int_name,
                                   bridge.br_name))


class SubPort(TrunkParentPort):
    """An OVS trunk subport.

    A subport is represented in OVS with two patch ports that
    connect a trunk bridge and the integration bridge respectively.
    These patch ports follow strict naming conventions: spi-<hash> for
    the patch port that goes into the integration bridge, and spt-<hash>
    for the patch port that goes into the trunk bridge.
    """
    DEV_PREFIX = 'sp'

    def __init__(self, trunk_id, port_id, port_mac=None, segmentation_id=None):
        super(SubPort, self).__init__(trunk_id, port_id, port_mac)
        self.segmentation_id = segmentation_id

    def plug(self, br_int):
        """Unplug patch ports between trunk bridge and given bridge.

        The method unplugs one patch port on the given bridge side using
        port MAC and ID as external IDs.  The other endpoint of patch port is
        attached to the trunk bridge.  Then it sets vlan tag represented by
        segmentation_id.  Everything is done in a single OVSDB transaction so
        either all operations succeed or fail.

        :param br_int: an integration bridge where peer endpoint of patch port
                       will be created.
        """
        ovsdb = self.bridge.ovsdb
        with ovsdb.transaction() as txn:
            super(SubPort, self).plug(br_int)
            txn.add(ovsdb.db_set(
                "Port", self.patch_port_trunk_name,
                ("tag", self.segmentation_id)))

    def unplug(self, bridge):
        """Unplug the sub port from the bridge.

        Method unplugs in single OVSDB transaction both endpoints of patch
        ports that represents the subport.

        :param bridge: bridge that has peer side of patch port for this
                       subport.
        """
        ovsdb = self.bridge.ovsdb
        with ovsdb.transaction() as txn:
            if self.bridge.exists():
                txn.add(ovsdb.del_port(self.patch_port_trunk_name,
                                       self.bridge.br_name))
            txn.add(ovsdb.del_port(self.patch_port_int_name,
                                   bridge.br_name))


class TrunkManager(object):
    """It implements the OVS trunk dataplane.

    It interfaces with the OVSDB server to execute OVS commands.
    """

    def __init__(self, br_int):
        self.br_int = br_int

    def create_trunk(self, trunk_id, port_id, port_mac):
        """Create the trunk.

        This patches the bridge for trunk_id with the integration bridge
        by means of parent port identified by port_id.

        :param trunk_id: ID of the trunk.
        :param port_id: ID of the parent port.
        :param port_mac: the MAC address of the parent port.
        :raises:
             TrunkBridgeNotFound: in case trunk bridge does not exist.
        """
        trunk = TrunkParentPort(trunk_id, port_id, port_mac)
        try:
            if not trunk.bridge.exists():
                raise exc.TrunkBridgeNotFound(bridge=trunk.bridge.br_name)
            trunk.plug(self.br_int)
        except RuntimeError as e:
            raise TrunkManagerError(error=e)

    def remove_trunk(self, trunk_id, port_id):
        """Remove the trunk bridge."""
        trunk = TrunkParentPort(trunk_id, port_id)
        try:
            trunk.unplug(self.br_int)
        except RuntimeError as e:
            raise TrunkManagerError(error=e)

    def dispose_trunk(self, trunk_bridge):
        """Clean up all the OVS resources associated to trunk_bridge."""
        ovsdb = trunk_bridge.ovsdb
        patch_peers = []
        try:
            patch_peers = trunk_bridge.get_ports_attributes(
                'Interface', columns=['options'])
            with trunk_bridge.ovsdb.transaction() as txn:
                for patch_peer in patch_peers:
                    peer_name = patch_peer['options'].get('peer')
                    if peer_name:
                        txn.add(ovsdb.del_port(peer_name, self.br_int.br_name))
                txn.add(ovsdb.del_br(trunk_bridge.br_name))
            LOG.debug("Deleted bridge '%s' and patch peers '%s'.",
                      trunk_bridge.br_name, patch_peers)
        except RuntimeError as e:
            LOG.error("Could not delete '%(peers)s' associated to "
                      "trunk bridge %(name)s. Reason: %(reason)s.",
                      {'peers': patch_peers,
                       'name': trunk_bridge.br_name,
                       'reason': e})

    def add_sub_port(self, trunk_id, port_id, port_mac, segmentation_id):
        """Create a sub_port.

        :param trunk_id: ID of the trunk.
        :param port_id: ID of the subport.
        :param segmentation_id: segmentation ID associated with this subport.
        :param port_mac: MAC address of the subport.
        """
        sub_port = SubPort(trunk_id, port_id, port_mac, segmentation_id)
        # If creating of parent trunk bridge takes longer than API call for
        # creating subport then bridge doesn't exist yet.
        try:
            if not sub_port.bridge.exists():
                raise exc.TrunkBridgeNotFound(bridge=sub_port.bridge.br_name)
            sub_port.plug(self.br_int)
        except RuntimeError as e:
            raise TrunkManagerError(error=e)

    def remove_sub_port(self, trunk_id, port_id):
        """Remove a sub_port.

        :param trunk_id: ID of the trunk.
        :param port_id: ID of the subport.
        """
        sub_port = SubPort(trunk_id, port_id)

        # Trunk bridge might have been deleted by calling delete_trunk() before
        # remove_sub_port().
        try:
            sub_port.unplug(self.br_int)
        except RuntimeError as e:
            raise TrunkManagerError(error=e)

    def get_port_uuid_from_external_ids(self, port):
        """Return the port UUID from the port metadata."""
        try:
            return self.br_int.portid_from_external_ids(
                port['external_ids'])
        except RuntimeError as e:
            raise TrunkManagerError(error=e)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Coordination and locking utilities."""

import inspect

import decorator
from oslo_concurrency import lockutils
from oslo_log import log
from oslo_utils import timeutils

LOG = log.getLogger(__name__)


def synchronized(lock_name):
    """Synchronization decorator.

    :param str lock_name: Lock name.

    Decorating a method like so::

        @synchronized('mylock')
        def foo(self, *args):
           ...

    ensures that only one process will execute the foo method at a time.

    Different methods can share the same lock::

        @synchronized('mylock')
        def foo(self, *args):
           ...

        @synchronized('mylock')
        def bar(self, *args):
           ...

    This way only one of either foo or bar can be executing at a time.

    Lock name can be formatted using Python format string syntax::

        @synchronized('{f_name}-{resource.id}-{snap[name]}')
        def foo(self, resource, snap):
           ...

    Available field names are: decorated function parameters and
    `f_name` as a decorated function name.
    """

    @decorator.decorator
    def _synchronized(f, *a, **k):
        sig = inspect.signature(f).bind(*a, **k)
        sig.apply_defaults()
        call_args = sig.arguments
        call_args['f_name'] = f.__name__
        lock_format_name = lock_name.format(**call_args)
        t1 = timeutils.now()
        t2 = None
        try:
            with lockutils.lock(lock_format_name):
                t2 = timeutils.now()
                LOG.debug('Lock "%(name)s" acquired by "%(function)s" :: '
                          'waited %(wait_secs)0.3fs',
                          {'name': lock_format_name,
                           'function': f.__name__,
                           'wait_secs': (t2 - t1)})
                return f(*a, **k)
        finally:
            t3 = timeutils.now()
            if t2 is None:
                held_secs = "N/A"
            else:
                held_secs = "%0.3fs" % (t3 - t2)
            LOG.debug('Lock "%(name)s" released by "%(function)s" :: held '
                      '%(held_secs)s',
                      {'name': lock_format_name,
                       'function': f.__name__,
                       'held_secs': held_secs})

    return _synchronized



# Copyright 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.agent import l3_extension
from neutron_lib import context
from oslo_utils import uuidutils

from neutron.agent.l3 import agent as l3_agent
from neutron.agent.l3 import l3_agent_extension_api as l3_ext_api
from neutron.agent.l3 import router_info as l3router
from neutron.api.rpc.callbacks import events
from neutron.services.logapi.agent.l3 import base as l3_base
from neutron.services.logapi.agent import log_extension as log_ext
from neutron.tests.unit.agent.l3 import test_agent

_uuid = uuidutils.generate_uuid


class FakeLogDriver(log_ext.LoggingDriver):

    SUPPORTED_LOGGING_TYPES = ['fake_resource']

    def initialize(self, resource_rpc, **kwargs):
        pass

    def start_logging(self, context, **kwargs):
        pass

    def stop_logging(self, context, **kwargs):
        pass


class FakeL3LoggingExtension(l3_base.L3LoggingExtensionBase,
                             l3_extension.L3AgentExtension):

    def initialize(self, connection, driver_type):
        pass


class L3LoggingExtBaseTestCase(test_agent.BasicRouterOperationsFramework):

    def setUp(self):
        super(L3LoggingExtBaseTestCase, self).setUp()
        self.agent = l3_agent.L3NATAgent('test_host', self.conf)
        self.context = context.get_admin_context()
        self.connection = mock.Mock()
        self.ex_gw_port = {'id': _uuid()}
        self.router = {'id': _uuid(),
                       'gw_port': self.ex_gw_port,
                       'ha': False,
                       'distributed': False}

        self.router_info = l3router.RouterInfo(self.agent, _uuid(),
                                               self.router, **self.ri_kwargs)
        self.router_info.ex_gw_port = self.ex_gw_port
        self.agent.router_info[self.router['id']] = self.router_info

        def _mock_get_router_info(router_id):
            return self.router_info

        self.get_router_info = mock.patch(
            'neutron.agent.l3.l3_agent_extension_api.'
            'L3AgentExtensionAPI.get_router_info').start()
        self.get_router_info.side_effect = _mock_get_router_info
        self.agent_api = l3_ext_api.L3AgentExtensionAPI(None, None)
        mock.patch(
            'neutron.manager.NeutronManager.load_class_for_provider').start()


class TestL3LoggingExtBase(L3LoggingExtBaseTestCase):

    def setUp(self):
        super(TestL3LoggingExtBase, self).setUp()
        self.agent_ext = FakeL3LoggingExtension()
        self.agent_ext.consume_api(self.agent_api)
        self.log_driver = mock.Mock()
        log_driver_object = FakeLogDriver()
        self.log_driver.defer_apply.side_effect = log_driver_object.defer_apply
        self.agent_ext.log_driver = self.log_driver

    def test__handle_notification_passes_update_events_enabled_log(self):
        log_obj = mock.Mock()
        log_obj.enabled = True
        self.agent_ext._handle_notification(
            self.context, 'log', [log_obj], events.UPDATED)
        self.assertTrue(self.log_driver.start_logging.called)

    def test__handle_notification_passes_update_events_disabled_log(self):
        log_obj = mock.Mock()
        log_obj.enabled = False
        self.agent_ext._handle_notification(
            self.context, 'log', [log_obj], events.UPDATED)
        self.assertTrue(self.log_driver.stop_logging.called)

    def test__handle_notification_passes_create_events(self):
        log_obj = mock.Mock()
        self.agent_ext._handle_notification(
            self.context, 'log', [log_obj], events.CREATED)
        self.assertTrue(self.log_driver.start_logging.called)

    def test__handle_notification_passes_delete_events(self):
        log_obj = mock.Mock()
        self.agent_ext._handle_notification(
            self.context, 'log', [log_obj], events.DELETED)
        self.assertTrue(self.log_driver.stop_logging.called)

    def test_add_router(self):
        self.agent_ext.add_router(self.context, self.router)
        self.log_driver.start_logging.assert_called_once_with(
            self.context, router_info=self.router_info)

    def test_update_router(self):
        self.agent_ext.update_router(self.context, self.router)
        self.log_driver.start_logging.assert_called_once_with(
            self.context, router_info=self.router_info)

    def test_delete_router(self):
        router_delete = {'id': _uuid()}
        self.agent_ext.delete_router(self.context, router_delete)
        self.log_driver.stop_logging.assert_called_once_with(
            self.context, router_info=router_delete)



# Copyright (C) 2014,2015 VA Linux Systems Japan K.K.
# Copyright (C) 2014,2015 YAMAMOTO Takashi <yamamoto at valinux co jp>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.plugins.ml2 import ovs_constants as ovs_const

from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent.openflow.native \
    import ovs_bridge_test_base


call = mock.call  # short hand


class OVSTunnelBridgeTest(ovs_bridge_test_base.OVSBridgeTestBase,
                          ovs_bridge_test_base.OVSDVRProcessTestMixin):
    dvr_process_table_id = ovs_const.DVR_PROCESS
    dvr_process_next_table_id = ovs_const.PATCH_LV_TO_TUN

    def setUp(self):
        conn_patcher = mock.patch(
            'neutron.agent.ovsdb.impl_idl._connection')
        conn_patcher.start()
        super(OVSTunnelBridgeTest, self).setUp()
        # NOTE(ivasilevskaya) The behaviour of oslotest.base.addCleanup()
        # according to https://review.opendev.org/#/c/119201/4 guarantees
        # that all started mocks will be stopped even without direct call to
        # patcher.stop().
        # If any individual mocks should be stopped by other than default
        # mechanism, their cleanup has to be added after
        # oslotest.BaseTestCase.setUp() not to be included in the stopall set
        # that will be cleaned up by mock.patch.stopall. This way the mock
        # won't be attempted to be stopped twice.
        self.addCleanup(conn_patcher.stop)
        self.setup_bridge_mock('br-tun', self.br_tun_cls)
        self.stamp = self.br.default_cookie

    def test_setup_default_table(self):
        patch_int_ofport = 5555
        arp_responder_enabled = False
        self.br.setup_default_table(patch_int_ofport=patch_int_ofport,
            arp_responder_enabled=arp_responder_enabled,
            dvr_enabled=False)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=2)],
                match=ofpp.OFPMatch(in_port=patch_int_ofport),
                priority=1, table_id=0),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=0),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=20)],
                match=ofpp.OFPMatch(
                    eth_dst=('00:00:00:00:00:00', '01:00:00:00:00:00')),
                priority=0,
                table_id=2),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=22)],
                match=ofpp.OFPMatch(
                    eth_dst=('01:00:00:00:00:00', '01:00:00:00:00:00')),
                priority=0,
                table_id=2),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=3),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=4),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=6),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.NXActionLearn(
                            cookie=self.stamp,
                            hard_timeout=300,
                            priority=1,
                            specs=[
                                ofpp.NXFlowSpecMatch(
                                    dst=('vlan_tci', 0),
                                    n_bits=12,
                                    src=('vlan_tci', 0)),
                                ofpp.NXFlowSpecMatch(
                                    dst=('eth_dst', 0),
                                    n_bits=48,
                                    src=('eth_src', 0)),
                                ofpp.NXFlowSpecLoad(
                                    dst=('vlan_tci', 0),
                                    n_bits=16,
                                    src=0),
                                ofpp.NXFlowSpecLoad(
                                    dst=('tunnel_id', 0),
                                    n_bits=64,
                                    src=('tunnel_id', 0)),
                                ofpp.NXFlowSpecOutput(
                                    dst='',
                                    n_bits=32,
                                    src=('in_port', 0)),
                            ],
                            table_id=20),
                        ofpp.OFPActionOutput(patch_int_ofport, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(),
                priority=1,
                table_id=10),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=22)],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=20),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=22),
                           active_bundle=None)
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_setup_default_table_arp_responder_enabled(self):
        patch_int_ofport = 5555
        arp_responder_enabled = True
        self.br.setup_default_table(patch_int_ofport=patch_int_ofport,
            arp_responder_enabled=arp_responder_enabled,
            dvr_enabled=False)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=2)],
                match=ofpp.OFPMatch(in_port=patch_int_ofport),
                priority=1, table_id=0),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=0),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=21)],
                match=ofpp.OFPMatch(
                    eth_dst='ff:ff:ff:ff:ff:ff',
                    eth_type=self.ether_types.ETH_TYPE_ARP),
                priority=1,
                table_id=2),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=20)],
                match=ofpp.OFPMatch(
                    eth_dst=('00:00:00:00:00:00', '01:00:00:00:00:00')),
                priority=0,
                table_id=2),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=22)],
                match=ofpp.OFPMatch(
                    eth_dst=('01:00:00:00:00:00', '01:00:00:00:00:00')),
                priority=0,
                table_id=2),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=3),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=4),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0, table_id=6),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.NXActionLearn(
                            cookie=self.stamp,
                            hard_timeout=300,
                            priority=1,
                            specs=[
                                ofpp.NXFlowSpecMatch(
                                    dst=('vlan_tci', 0),
                                    n_bits=12,
                                    src=('vlan_tci', 0)),
                                ofpp.NXFlowSpecMatch(
                                    dst=('eth_dst', 0),
                                    n_bits=48,
                                    src=('eth_src', 0)),
                                ofpp.NXFlowSpecLoad(
                                    dst=('vlan_tci', 0),
                                    n_bits=16,
                                    src=0),
                                ofpp.NXFlowSpecLoad(
                                    dst=('tunnel_id', 0),
                                    n_bits=64,
                                    src=('tunnel_id', 0)),
                                ofpp.NXFlowSpecOutput(
                                    dst='',
                                    n_bits=32,
                                    src=('in_port', 0)),
                            ],
                            table_id=20),
                        ofpp.OFPActionOutput(patch_int_ofport, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(),
                priority=1,
                table_id=10),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=22)],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=20),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=22)],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=21),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=22),
                           active_bundle=None)
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def _test_setup_default_table_dvr_helper(self, dvr_enabled):
        patch_int_ofport = 5555
        arp_responder_enabled = True
        self.br.setup_default_table(patch_int_ofport=patch_int_ofport,
            arp_responder_enabled=arp_responder_enabled,
            dvr_enabled=dvr_enabled)
        (dp, ofp, ofpp) = self._get_dp()
        non_dvr_specific_call = call._send_msg(
            ofpp.OFPFlowMod(
                dp,
                cookie=self.stamp,
                instructions=[ofpp.OFPInstructionGotoTable(table_id=2)],
                match=ofpp.OFPMatch(in_port=patch_int_ofport),
                priority=1, table_id=0),
            active_bundle=None)

        if dvr_enabled:
            self.assertNotIn(non_dvr_specific_call, self.mock.mock_calls)
        else:
            self.assertIn(non_dvr_specific_call, self.mock.mock_calls)

    def test_setup_default_table_dvr_enabled(self):
        self._test_setup_default_table_dvr_helper(dvr_enabled=True)

    def test_setup_default_table_dvr_disabled(self):
        self._test_setup_default_table_dvr_helper(dvr_enabled=False)

    def test_provision_local_vlan(self):
        network_type = 'vxlan'
        lvid = 888
        segmentation_id = 777
        distributed = False
        self.br.provision_local_vlan(network_type=network_type, lvid=lvid,
                                     segmentation_id=segmentation_id,
                                     distributed=distributed)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPushVlan(),
                        ofpp.OFPActionSetField(
                            vlan_vid=lvid | ofp.OFPVID_PRESENT)
                    ]),
                    ofpp.OFPInstructionGotoTable(table_id=10),
                ],
                match=ofpp.OFPMatch(tunnel_id=segmentation_id),
                priority=1,
                table_id=4),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_reclaim_local_vlan(self):
        network_type = 'vxlan'
        segmentation_id = 777
        self.br.reclaim_local_vlan(network_type=network_type,
                                   segmentation_id=segmentation_id)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(
                table_id=4,
                match=ofpp.OFPMatch(tunnel_id=segmentation_id)),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_install_flood_to_tun(self):
        vlan = 3333
        tun_id = 2222
        ports = [11, 44, 22, 33]
        self.br.install_flood_to_tun(vlan=vlan,
                                     tun_id=tun_id,
                                     ports=ports)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPopVlan(),
                        ofpp.OFPActionSetField(tunnel_id=tun_id),
                    ] + [ofpp.OFPActionOutput(p, 0) for p in ports]),
                ],
                match=ofpp.OFPMatch(vlan_vid=vlan | ofp.OFPVID_PRESENT),
                priority=1,
                table_id=22),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_flood_to_tun(self):
        vlan = 3333
        self.br.delete_flood_to_tun(vlan=vlan)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(table_id=22,
                match=ofpp.OFPMatch(vlan_vid=vlan | ofp.OFPVID_PRESENT)),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_install_unicast_to_tun(self):
        vlan = 3333
        port = 55
        mac = '08:60:6e:7f:74:e7'
        tun_id = 2222
        self.br.install_unicast_to_tun(vlan=vlan,
                                       tun_id=tun_id,
                                       port=port,
                                       mac=mac)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPopVlan(),
                        ofpp.OFPActionSetField(tunnel_id=tun_id),
                        ofpp.OFPActionOutput(port, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    eth_dst=mac, vlan_vid=vlan | ofp.OFPVID_PRESENT),
                priority=2,
                table_id=20),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_unicast_to_tun(self):
        vlan = 3333
        mac = '08:60:6e:7f:74:e7'
        self.br.delete_unicast_to_tun(vlan=vlan, mac=mac)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(table_id=20,
                match=ofpp.OFPMatch(
                    eth_dst=mac, vlan_vid=vlan | ofp.OFPVID_PRESENT)),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_unicast_to_tun_without_mac(self):
        vlan = 3333
        mac = None
        self.br.delete_unicast_to_tun(vlan=vlan, mac=mac)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(table_id=20,
                match=ofpp.OFPMatch(vlan_vid=vlan | ofp.OFPVID_PRESENT)),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_arp_responder(self):
        vlan = 3333
        ip = '192.0.2.1'
        self.br.delete_arp_responder(vlan=vlan, ip=ip)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(
                match=ofpp.OFPMatch(
                    eth_type=self.ether_types.ETH_TYPE_ARP,
                    arp_tpa=ip,
                    vlan_vid=vlan | ofp.OFPVID_PRESENT),
                table_id=21),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_arp_responder_without_ip(self):
        vlan = 3333
        ip = None
        self.br.delete_arp_responder(vlan=vlan, ip=ip)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(
                match=ofpp.OFPMatch(
                    eth_type=self.ether_types.ETH_TYPE_ARP,
                    vlan_vid=vlan | ofp.OFPVID_PRESENT),
                table_id=21),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_setup_tunnel_port(self):
        network_type = 'vxlan'
        port = 11111
        self.br.setup_tunnel_port(network_type=network_type, port=port)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionGotoTable(table_id=4),
                ],
                match=ofpp.OFPMatch(in_port=port),
                priority=1,
                table_id=0),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_cleanup_tunnel_port(self):
        port = 11111
        self.br.cleanup_tunnel_port(port=port)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(in_port=port),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_add_dvr_mac_tun(self):
        mac = '00:02:b3:13:fe:3d'
        port = 8888
        self.br.add_dvr_mac_tun(mac=mac, port=port)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionOutput(port, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(eth_src=mac),
                priority=1,
                table_id=9),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_remove_dvr_mac_tun(self):
        mac = '00:02:b3:13:fe:3d'
        self.br.remove_dvr_mac_tun(mac=mac)
        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(eth_src=mac, table_id=9),
        ]
        self.assertEqual(expected, self.mock.mock_calls)



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures
from neutron_lib.utils import runtime
from oslo_log import log as logging
from oslo_utils import fileutils


LOG = logging.getLogger(__name__)
MAX_ATTEMPTS = 100
TMP_DIR = '/tmp/neutron_exclusive_resources/'


class ExclusiveResource(fixtures.Fixture):
    def __init__(self, resource_name, allocator_function, validator=None):
        self.ra = ResourceAllocator(
            resource_name, allocator_function, validator)

    def _setUp(self):
        self.resource = self.ra.allocate()
        self.addCleanup(self.ra.release, self.resource)


class ResourceAllocator(object):
    """ResourceAllocator persists cross-process allocations of a resource.

    Allocations are persisted to a file determined by the 'resource_name',
    and are allocated via an allocator_function. The public interface
    (allocate and release) are guarded by a file lock. The intention
    is to allow atomic, cross-process allocation of shared resources
    such as ports and IP addresses. For usages of this class, please see
    ExclusiveIPAddress and its functional tests.

    Note that this class doesn't maintain in-memory state, and multiple
    instances of it may be initialized and used. A pool of resources
    is identified solely by the 'resource_name' argument.
    """
    def __init__(self, resource_name, allocator_function, validator=None):
        """Initialize a resource allocator.

        :param resource_name: A unique identifier for a pool of resources.
        :param allocator_function: A function with no parameters that generates
                                   a resource.
        :param validator: An optional function that accepts a resource and an
                          existing pool and returns if the generated resource
                          is valid.
        """
        def is_valid(new_resource, allocated_resources):
            return new_resource not in allocated_resources

        self._allocator_function = allocator_function
        self._state_file_path = os.path.join(TMP_DIR, resource_name)
        self._validator = validator if validator else is_valid
        self._resource_name = resource_name

    @runtime.synchronized('resource_allocator', external=True,
                          lock_path='/tmp')
    def allocate(self):
        allocations = self._get_allocations()

        for i in range(MAX_ATTEMPTS):
            resource = str(self._allocator_function())
            if self._validator(resource, allocations):
                allocations.add(resource)
                self._write_allocations(allocations)
                LOG.debug('Allocated exclusive resource %s of type %s. '
                          'The allocations are now: %s',
                          resource, self._resource_name, allocations)
                return resource

        raise ValueError(
            'Could not allocate a new resource of type %s from pool %s' %
            (self._resource_name, allocations))

    @runtime.synchronized('resource_allocator', external=True,
                          lock_path='/tmp')
    def release(self, resource):
        allocations = self._get_allocations()
        allocations.remove(resource)
        if allocations:
            self._write_allocations(allocations)
        else:  # Clean up the file if we're releasing the last allocation
            os.remove(self._state_file_path)
        LOG.debug('Released exclusive resource %s of type %s. The allocations '
                  'are now: %s',
                  resource, self._resource_name, allocations)

    def _get_allocations(self):
        fileutils.ensure_tree(TMP_DIR, mode=0o755)

        try:
            with open(self._state_file_path, 'r') as allocations_file:
                contents = allocations_file.read()
        except IOError:
            contents = None

        # If the file was empty, we want to return an empty set, not {''}
        return set(contents.split(',')) if contents else set()

    def _write_allocations(self, allocations):
        with open(self._state_file_path, 'w') as allocations_file:
            allocations_file.write(','.join(allocations))



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
from unittest import mock

from neutron_lib import constants
from oslo_config import fixture as fixture_config
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import event
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.tests.functional.schema.ovn_southbound import event as test_event
import testtools

from neutron.agent.linux import iptables_manager
from neutron.agent.ovn.metadata import agent
from neutron.agent.ovn.metadata import ovsdb
from neutron.agent.ovn.metadata import server as metadata_server
from neutron.common.ovn import constants as ovn_const
from neutron.common import utils as n_utils
from neutron.conf.agent.metadata import config as meta_config
from neutron.conf.agent.ovn.metadata import config as meta_config_ovn
from neutron.tests.common import net_helpers
from neutron.tests.functional import base
from neutron.tests.functional.common import ovn as ovn_common

AGENT_CHASSIS_TABLE = 'Chassis_Private'


class NoDatapathProvision(Exception):
    pass


class MetadataAgentHealthEvent(event.WaitEvent):
    event_name = 'MetadataAgentHealthEvent'

    def __init__(self, chassis, sb_cfg, timeout=5):
        self.chassis = chassis
        self.sb_cfg = sb_cfg
        super(MetadataAgentHealthEvent, self).__init__(
            (self.ROW_UPDATE,),
            AGENT_CHASSIS_TABLE,
            (('name', '=', self.chassis),),
            timeout=timeout)

    def matches(self, event, row, old=None):
        if not super(MetadataAgentHealthEvent, self).matches(event, row, old):
            return False
        return int(row.external_ids.get(
            ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY, 0)) >= self.sb_cfg


class MetadataPortCreateEvent(event.WaitEvent):
    event_name = 'MetadataPortCreateEvent'

    def __init__(self, metadata_port, timeout=5):
        table = 'Port_Binding'
        events = (self.ROW_CREATE,)
        conditions = (('logical_port', '=', metadata_port),)
        super(MetadataPortCreateEvent, self).__init__(
            events, table, conditions, timeout=timeout
        )


class TestMetadataAgent(base.TestOVNFunctionalBase):
    OVN_BRIDGE = 'br-int'
    FAKE_CHASSIS_HOST = 'ovn-host-fake'

    def setUp(self):
        super(TestMetadataAgent, self).setUp()
        self.handler = self.sb_api.idl.notify_handler
        # We only have OVN NB and OVN SB running for functional tests
        self.mock_ovsdb_idl = mock.Mock()
        mock_metadata_instance = mock.Mock()
        mock_metadata_instance.start.return_value = self.mock_ovsdb_idl
        mock_metadata = mock.patch.object(
            ovsdb, 'MetadataAgentOvsIdl').start()
        mock_metadata.return_value = mock_metadata_instance
        self._mock_get_ovn_br = mock.patch.object(
            agent.MetadataAgent,
            '_get_ovn_bridge',
            return_value=self.OVN_BRIDGE).start()
        self.agent = self._start_metadata_agent()

    def _start_metadata_agent(self):
        conf = self.useFixture(fixture_config.Config()).conf
        conf.register_opts(meta_config.SHARED_OPTS)
        conf.register_opts(meta_config.UNIX_DOMAIN_METADATA_PROXY_OPTS)
        conf.register_opts(meta_config.METADATA_PROXY_HANDLER_OPTS)
        conf.register_opts(meta_config_ovn.OVS_OPTS, group='ovs')
        meta_config_ovn.setup_privsep()

        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        conf.set_override('ovn_sb_connection', ovn_sb_db, group='ovn')
        conf.set_override('metadata_workers', '0')

        self.chassis_name = self.add_fake_chassis(self.FAKE_CHASSIS_HOST)
        mock.patch.object(agent.MetadataAgent,
                          '_get_own_chassis_name',
                          return_value=self.chassis_name).start()
        agt = agent.MetadataAgent(conf)
        with mock.patch.object(metadata_server.UnixDomainMetadataProxy,
                               'wait'):
            agt.start()

        # Metadata agent will open connections to OVS and SB databases.
        # Close connections to them when the test ends,
        self.addCleanup(agt.ovs_idl.ovsdb_connection.stop)
        self.addCleanup(agt.sb_idl.ovsdb_connection.stop)

        return agt

    def test_metadata_agent_healthcheck(self):
        chassis_row = self.sb_api.db_find(
            AGENT_CHASSIS_TABLE,
            ('name', '=', self.chassis_name)).execute(
            check_error=True)[0]

        # Assert that, prior to creating a resource the metadata agent
        # didn't populate the external_ids from the Chassis
        self.assertNotIn(ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY,
                         chassis_row['external_ids'])

        # Let's list the agents to force the nb_cfg to be bumped on NB
        # db, which will automatically increment the nb_cfg counter on
        # NB_Global and make ovn-controller copy it over to SB_Global. Upon
        # this event, Metadata agent will update the external_ids on its
        # Chassis row to signal that it's healthy.

        row_event = MetadataAgentHealthEvent(self.chassis_name, 1)
        self.handler.watch_event(row_event)
        self.new_list_request('agents').get_response(self.api)

        # If we do not time out waiting for the event, then we are assured
        # that the metadata agent has populated the external_ids from the
        # chassis with the nb_cfg, 1 revisions when listing the agents.
        self.assertTrue(row_event.wait())

    def _create_metadata_port(self, txn, lswitch_name, port_name=None):
        mdt_port_name = port_name or 'ovn-mdt-' + uuidutils.generate_uuid()
        txn.add(
            self.nb_api.lsp_add(
                lswitch_name,
                mdt_port_name,
                type=ovn_const.LSP_TYPE_LOCALPORT,
                addresses='AA:AA:AA:AA:AA:AA 192.168.122.123',
                external_ids={
                    ovn_const.OVN_CIDRS_EXT_ID_KEY: '192.168.122.123/24',
                    ovn_const.OVN_DEVID_EXT_ID_KEY:
                        ovn_const.OVN_METADATA_PREFIX + lswitch_name,
                    ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                        constants.DEVICE_OWNER_DISTRIBUTED
                }))

    def _update_metadata_port_ip(self, metadata_port_name):
        external_ids = {
            ovn_const.OVN_CIDRS_EXT_ID_KEY: "192.168.122.2/24",
            ovn_const.OVN_DEVID_EXT_ID_KEY:
                ovn_const.OVN_METADATA_PREFIX + uuidutils.generate_uuid()
        }
        self.nb_api.set_lswitch_port(lport_name=metadata_port_name,
                                     external_ids=external_ids).execute()

    def _create_logical_switch_port(self, type_=None, addresses=None):
        lswitch_name = 'ovn-' + uuidutils.generate_uuid()
        lswitchport_name = 'ovn-port-' + uuidutils.generate_uuid()
        # It may take some time to ovn-northd to translate from OVN NB DB to
        # the OVN SB DB. Wait for port binding event to happen before binding
        # the port to chassis.
        pb_event = test_event.WaitForPortBindingEvent(lswitchport_name)
        self.handler.watch_event(pb_event)

        lswitch_port_columns = {}
        if addresses:
            lswitch_port_columns['addresses'] = addresses
        if type_:
            lswitch_port_columns['type'] = type_

        with self.nb_api.transaction(check_error=True, log_errors=True) as txn:
            txn.add(
                self.nb_api.ls_add(lswitch_name))
            txn.add(
                self.nb_api.create_lswitch_port(
                    lswitchport_name, lswitch_name, **lswitch_port_columns))
            self._create_metadata_port(txn, lswitch_name)
        self.assertTrue(pb_event.wait())

        return lswitchport_name, lswitch_name

    def test_agent_resync_on_non_existing_bridge(self):
        BR_NEW = 'br-new'
        self._mock_get_ovn_br.return_value = BR_NEW
        self.agent.ovs_idl.list_br.return_value.execute.return_value = [BR_NEW]
        # The agent has initialized with br-int and above list_br doesn't
        # return it, hence the agent should trigger reconfiguration and store
        # new br-new value to its attribute.
        self.assertEqual(self.OVN_BRIDGE, self.agent.ovn_bridge)

        # NOTE: The IP address is specifically picked such that it fits the
        # metadata port external_ids: { neutron:cidrs }. This is because agent
        # will only trigger if the logical port is part of a neutron subnet
        lswitchport_name, _ = self._create_logical_switch_port(
            addresses='AA:AA:AA:AA:AA:AB 192.168.122.125'
        )

        # Trigger PortBindingChassisCreatedEvent
        self.sb_api.lsp_bind(lswitchport_name, self.chassis_name).execute(
            check_error=True, log_errors=True)

        exc = Exception("Agent bridge hasn't changed from %s to %s "
                        "in 10 seconds after Port_Binding event" %
                        (self.agent.ovn_bridge, BR_NEW))
        n_utils.wait_until_true(
            lambda: BR_NEW == self.agent.ovn_bridge,
            timeout=10,
            exception=exc)

    def _test_agent_events_prepare(self, lsp_type=None):
        lswitchport_name, lswitch_name = self._create_logical_switch_port(
            lsp_type)
        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:
            self.sb_api.lsp_bind(lswitchport_name, self.chassis_name).execute(
                check_error=True, log_errors=True)

            # Wait until port is bound
            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=Exception(
                    "Datapath provisioning did not happen on port binding"))

        return lswitchport_name, lswitch_name

    def test_agent_unbind_port(self):
        lswitchport_name, lswitch_name = self._test_agent_events_prepare()

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:
            self.sb_api.lsp_unbind(lswitchport_name).execute(
                check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=Exception(
                    "Datapath teardown did not happen after the port was "
                    "unbound"))

    def _test_agent_delete_bound_external_port(self, lsp_type=None):
        lswitchport_name, lswitch_name = self._test_agent_events_prepare(
            lsp_type)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision,\
                mock.patch.object(agent.LOG, 'warning') as m_log_warn:
            self.nb_api.delete_lswitch_port(
                lswitchport_name, lswitch_name).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=Exception(
                    "Datapath teardown did not happen after external port was "
                    "deleted"))
            if lsp_type == ovn_const.LSP_TYPE_EXTERNAL:
                m_log_warn.assert_not_called()
            else:
                m_log_warn.assert_called()

    def test_agent_delete_bound_external_port(self):
        self._test_agent_delete_bound_external_port(
            lsp_type=ovn_const.LSP_TYPE_EXTERNAL)

    def test_agent_delete_bound_nonexternal_port(self):
        self._test_agent_delete_bound_external_port()

    def test_agent_registration_at_chassis_create_event(self):
        def check_for_metadata():
            chassis = self.sb_api.lookup(
                AGENT_CHASSIS_TABLE, self.chassis_name)
            return ovn_const.OVN_AGENT_METADATA_ID_KEY in chassis.external_ids

        exc = Exception('Chassis not created, %s is not in chassis '
                        'external_ids' % ovn_const.OVN_AGENT_METADATA_ID_KEY)
        n_utils.wait_until_true(check_for_metadata, timeout=5, exception=exc)

        # Delete Chassis and assert
        chassis = self.sb_api.lookup('Chassis', self.chassis_name)
        self.del_fake_chassis(chassis.name)
        self.assertRaises(idlutils.RowNotFound, self.sb_api.lookup,
                          'Chassis', self.chassis_name)

        # Re-add the Chassis
        self.add_fake_chassis(self.FAKE_CHASSIS_HOST, name=self.chassis_name)
        exc = Exception('Agent metadata failed to re-register itself '
                        'after the Chassis %s was re-created' %
                        self.chassis_name)

        # Check if metadata agent was re-registered
        chassis = self.sb_api.lookup('Chassis', self.chassis_name)
        n_utils.wait_until_true(
            check_for_metadata,
            timeout=10,
            exception=exc)

    def test_agent_metadata_port_ip_update_event(self):
        lswitch_name = 'ovn-' + uuidutils.generate_uuid()
        mdt_port_name = 'ovn-mdt-' + uuidutils.generate_uuid()

        mdt_pb_event = test_event.WaitForPortBindingEvent(mdt_port_name)
        self.handler.watch_event(mdt_pb_event)

        with self.nb_api.transaction(
                check_error=True, log_errors=True) as txn:
            txn.add(
                self.nb_api.ls_add(lswitch_name))
            self._create_metadata_port(txn, lswitch_name, mdt_port_name)

        self.assertTrue(mdt_pb_event.wait())

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:
            self.sb_api.lsp_bind(mdt_port_name, self.chassis_name).execute(
                check_error=True, log_errors=True)

            # Wait until port is bound
            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=Exception(
                    "Datapath provisioning did not happen on port binding"))

            m_provision.reset_mock()

            self._update_metadata_port_ip(mdt_port_name)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=Exception(
                    "Datapath provisioning not called after external ids was "
                    "changed"))

    def test_metadata_agent_only_monitors_own_chassis(self):
        # We already have the fake chassis which we should be monitoring, so
        # create an event looking for a change to another chassis
        other_name = uuidutils.generate_uuid()
        other_chassis = self.add_fake_chassis(self.FAKE_CHASSIS_HOST,
                                              name=other_name)
        self.assertEqual(other_chassis, other_name)

        event = MetadataAgentHealthEvent(chassis=other_name, sb_cfg=-1,
                                         timeout=0)
        # Use the agent's sb_idl to watch for the event since it has condition
        self.agent.sb_idl.idl.notify_handler.watch_event(event)
        # Use the test sb_api to set other_chassis values since shouldn't exist
        # on agent's sb_idl
        self.sb_api.db_set(
            AGENT_CHASSIS_TABLE, other_chassis,
            ('external_ids', {'test': 'value'})).execute(check_error=True)

        event2 = MetadataAgentHealthEvent(chassis=self.chassis_name, sb_cfg=-1)
        self.agent.sb_idl.idl.notify_handler.watch_event(event2)
        # Use the test's sb_api again to send a command so we can see if it
        # completes and short-circuit the need to wait for a timeout to pass
        # the test. If we get the result to this, we would have gotten the
        # previous result as well.
        self.sb_api.db_set(
            AGENT_CHASSIS_TABLE, self.chassis_name,
            ('external_ids', {'test': 'value'})).execute(check_error=True)
        self.assertTrue(event2.wait())
        self.assertFalse(event.wait())

    def test__ensure_datapath_checksum_if_dpdk(self):
        self.mock_ovsdb_idl.db_get.return_value.execute.return_value = (
            ovn_const.CHASSIS_DATAPATH_NETDEV)
        regex = re.compile(r'-A POSTROUTING -p tcp -m tcp '
                           r'-j CHECKSUM --checksum-fill')
        namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        self.agent._ensure_datapath_checksum(namespace)
        iptables_mgr = iptables_manager.IptablesManager(
            use_ipv6=True, nat=False, namespace=namespace, external_lock=False)
        for rule in iptables_mgr.get_rules_for_table('mangle'):
            if regex.match(rule):
                return
        self.fail('Rule not found in "mangle" table, in namespace %s' %
                  namespace)

    def test_metadata_proxy_handler_idl(self):
        # This test relies on the configuration option metadata_workers=0
        proxy_sb_idl = self.agent._proxy.server._server._application.sb_idl
        agent_sb_idl = self.agent.sb_idl
        self.assertEqual(agent_sb_idl, proxy_sb_idl)

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_provisioned_on_additional_chassis_change(self):
        other_chassis_name = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis", name=other_chassis_name)

        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, other_chassis_name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            # Update the additional_chassis
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[agent_chassis.uuid]).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Additional chassis didn't trigger Port Binding event"))

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_not_provisioned_on_foreign_additional_chassis_change(
            self):
        other_chassis_name = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis", name=other_chassis_name)

        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)
        other_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', other_chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, agent_chassis.name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            # Update the additional_chassis, the agent should not see the
            # notification because it has only its own chassis row locally and
            # does not see other chassis
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[other_chassis.uuid]).execute(
                    check_error=True, log_errors=True)

            with testtools.ExpectedException(NoDatapathProvision):
                n_utils.wait_until_true(
                    lambda: m_provision.called,
                    timeout=1,
                    exception=NoDatapathProvision(
                        "Provisioning wasn't triggered"))

    @ovn_common.skip_if_additional_chassis_not_supported
    def test_metadata_teardown_on_additional_chassis_removed(self):
        other_chassis_name = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis", name=other_chassis_name)

        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, other_chassis_name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            # Update the additional_chassis
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[agent_chassis.uuid]).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Additional chassis didn't trigger Port Binding event"))

            m_provision.reset_mock()

            # Remove the additional_chassis but keep the chassis. This is
            # simulates the live migration has failed
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[]).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Removing additional chassis did not call teardown"))

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_additional_chassis_removed_chassis_set(self):
        other_chassis_name = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis", name=other_chassis_name)

        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, other_chassis_name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            # Update the additional_chassis
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[agent_chassis.uuid]).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Additional chassis didn't trigger Port Binding event"))

            m_provision.reset_mock()

            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[], chassis=agent_chassis.uuid).execute(
                    check_error=True, log_errors=True)

            with testtools.ExpectedException(NoDatapathProvision):
                n_utils.wait_until_true(
                    lambda: m_provision.called,
                    timeout=1,
                    exception=NoDatapathProvision(
                        "Removing additional chassis did not call teardown"))

    def _test_metadata_additional_chassis_removed(self, new_chassis_uuid):
        other_chassis_name = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis", name=other_chassis_name)

        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, other_chassis_name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            # Update the additional_chassis
            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[agent_chassis.uuid]).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Additional chassis didn't trigger Port Binding event"))

            m_provision.reset_mock()

            self.sb_api.db_set('Port_Binding', pb.uuid,
                additional_chassis=[], chassis=new_chassis_uuid).execute(
                    check_error=True, log_errors=True)

            n_utils.wait_until_true(
                lambda: m_provision.called,
                timeout=10,
                exception=NoDatapathProvision(
                    "Removing additional chassis did not call teardown"))

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_additional_chassis_removed_different_chassis_set(self):
        other_chassis_name2 = uuidutils.generate_uuid()
        self.add_fake_chassis("other_chassis2", name=other_chassis_name2)
        other_chassis2 = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', other_chassis_name2)
        self._test_metadata_additional_chassis_removed(other_chassis2.uuid)

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_additional_chassis_removed_chassis_unset(self):
        self._test_metadata_additional_chassis_removed(new_chassis_uuid=[])

    @ovn_common.skip_if_additional_chassis_not_supported('sb_api')
    def test_metadata_port_binding_column_updated(self):
        agent_chassis = idlutils.row_by_value(
            self.sb_api, 'Chassis', 'name', self.chassis_name)

        lswitchport_name, lswitch_name = self._create_logical_switch_port()

        self.sb_api.lsp_bind(
            lswitchport_name, agent_chassis.name).execute(
                check_error=True, log_errors=True)
        pb = idlutils.row_by_value(
            self.sb_api, 'Port_Binding', 'logical_port', lswitchport_name)

        with mock.patch.object(
                agent.MetadataAgent, 'provision_datapath') as m_provision:

            self.sb_api.db_add('Port_Binding', pb.uuid,
                'external_ids', {'foo': 'bar'}).execute(
                    check_error=True, log_errors=True)

            with testtools.ExpectedException(NoDatapathProvision):
                n_utils.wait_until_true(
                    lambda: m_provision.called,
                    timeout=1,
                    exception=NoDatapathProvision(
                        "Provisioning wasn't triggered"))



# Copyright 2021 Troila
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.agent.l3 import agent as neutron_l3_agent
from neutron.agent.l3.extensions import ndp_proxy as np
from neutron.agent.linux import ip_lib
from neutron.agent.linux import iptables_manager as iptable_mng
from neutron.api.rpc.callbacks import resources
from neutron.common import utils as common_utils
from neutron.objects import agent as agent_obj
from neutron.objects import ndp_proxy as np_obj
from neutron.objects import ports as ports_obj
from neutron.tests.functional.agent.l3 import framework
from neutron.tests.functional.agent.l3 import test_dvr_router

HOSTNAME = 'agent1'


class L3AgentNDPProxyTestFramework(framework.L3AgentTestFramework):

    def setUp(self):
        super(L3AgentNDPProxyTestFramework, self).setUp()
        self.conf.set_override('extensions', ['ndp_proxy'], 'agent')
        self.agent = neutron_l3_agent.L3NATAgentWithStateReport(HOSTNAME,
                                                                self.conf)
        self.np_ext = np.NDPProxyAgentExtension()

        port_id1 = uuidutils.generate_uuid()
        port1_binding = ports_obj.PortBinding(port_id=port_id1,
                                              host=self.agent.host)
        port1_obj = ports_obj.Port(id=port_id1, bindings=[port1_binding])
        port_id2 = uuidutils.generate_uuid()
        port2_binding = ports_obj.PortBinding(port_id=port_id1,
                                              host='fake_host')
        port2_obj = ports_obj.Port(id=port_id2, bindings=[port2_binding])
        self.ports = [port1_obj, port2_obj]
        self.port_binding_map = {port_id1: port1_binding,
                                 port_id2: port2_binding}
        self.ndpproxy1 = np_obj.NDPProxy(
            context=None, id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            port_id=port_id1, ip_address='2002::1:3')
        self.ndpproxy2 = np_obj.NDPProxy(
            context=None, id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            port_id=port_id2, ip_address='2002::1:4')
        self.ndp_proxies = [self.ndpproxy1, self.ndpproxy2]
        agent_configurations = {
            'agent_mode': constants.L3_AGENT_MODE_DVR_NO_EXTERNAL}
        self.agent_obj = agent_obj.Agent(
            id=uuidutils.generate_uuid(), host=self.agent.host,
            agent_type=constants.AGENT_TYPE_L3,
            configurations=agent_configurations)
        self._set_pull_mock()

    def _set_pull_mock(self):

        def _bulk_pull_mock(context, resource_type, filter_kwargs=None):
            if resource_type == resources.PORT:
                return [port for port in self.ports if
                        port.id == filter_kwargs['id']]
            if resource_type == resources.AGENT:
                return [self.agent_obj]
            if resource_type == resources.NDPPROXY:
                result = []
                if 'router_id' in filter_kwargs:
                    for ndp_proxy in self.ndp_proxies:
                        if ndp_proxy.router_id in filter_kwargs['router_id']:
                            result.append(ndp_proxy)
                    return result
                return self.ndp_proxie

        self.pull = mock.patch('neutron.api.rpc.handlers.resources_rpc.'
                               'ResourcesPullRpcApi.pull').start()
        self.bulk_pull = mock.patch('neutron.api.rpc.handlers.resources_rpc.'
                                    'ResourcesPullRpcApi.bulk_pull').start()
        self.bulk_pull.side_effect = _bulk_pull_mock

    def _get_existing_ndp_proxies(self, interface_name, namespace):
        ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
        cmd = ['ip', '-6', 'neigh', 'list', 'proxy']
        res = ip_wrapper.netns.execute(cmd)
        proxies = []
        for proxy in res.split('\n'):
            # Exclude null line
            if proxy:
                proxy_list = proxy.split(' ')
                if interface_name in proxy_list:
                    try:
                        if netaddr.IPAddress(proxy_list[0]).version == 6:
                            proxies.append(proxy_list[0])
                    except Exception:
                        pass
        return proxies

    def _assert_ndp_proxy_kernel_parameter(self, ip_wrapper, interface_name):
        sysctl_cmd = ['sysctl', '-b',
                      'net.ipv6.conf.%s.proxy_ndp' % interface_name]

        def check_kernel_parameter():
            res = ip_wrapper.netns.execute(sysctl_cmd, privsep_exec=True)
            if res == "1":
                return True

        common_utils.wait_until_true(check_kernel_parameter)

    def _assert_ndp_iptable_chain_is_set(self, iptables_manager,
                                         interface_name):
        rule = '-i %s -j DROP' % interface_name
        rule_obj = iptable_mng.IptablesRule('NDP', rule, True, False,
                                            iptables_manager.wrap_name)

        def check_chain_is_set():
            existing_chains = iptables_manager.ipv6['filter'].chains
            if 'NDP' not in existing_chains:
                return False

            existing_rules = iptables_manager.ipv6['filter'].rules
            if rule_obj in existing_rules:
                return True

        common_utils.wait_until_true(check_chain_is_set)

    def _assert_ndp_proxy_state_iptable_rules_is_set(
            self, ri, iptables_manager, interface_name):
        wrap_name = iptables_manager.wrap_name
        expected_rules = []
        for port in ri.internal_ports:
            for subnet in port['subnets']:
                if netaddr.IPNetwork(subnet['cidr']).version == \
                        constants.IP_VERSION_4:
                    continue
                rule = (
                    '-i %s --destination %s -j '
                    '%s-NDP') % (interface_name,
                                 subnet['cidr'],
                                 wrap_name)

                rule_obj = iptable_mng.IptablesRule(
                    'FORWARD', rule, True, False, iptables_manager.wrap_name)
                expected_rules.append(rule_obj)

        def check_rules_is_set():
            existing_rules = iptables_manager.ipv6['filter'].rules
            for rule in expected_rules:
                if rule not in existing_rules:
                    return False
            return True

        common_utils.wait_until_true(check_rules_is_set)

    def _assect_ndp_proxy_rules_is_set(self, ip_wrapper, iptables_manager,
                                       interface_name, namespace):
        expected_iptable_rules = []
        expected_proxy_address = []
        for ndp_proxy in self.ndp_proxies:
            rule = '-i %s --destination %s -j ACCEPT' % (interface_name,
                                                         ndp_proxy.ip_address)
            rule_obj = iptable_mng.IptablesRule('NDP', rule, True, True,
                                                iptables_manager.wrap_name)
            expected_iptable_rules.append(rule_obj)
            expected_proxy_address.append(str(ndp_proxy.ip_address))

        def check_rules_is_set():
            existing_iptable_rules = iptables_manager.ipv6['filter'].rules
            for iptable_rule in expected_iptable_rules:
                if iptable_rule not in existing_iptable_rules:
                    return False

            existing_proxy_addresses = self._get_existing_ndp_proxies(
                interface_name, namespace)
            for address in expected_proxy_address:
                if address not in existing_proxy_addresses:
                    return False

            return True

        common_utils.wait_until_true(check_rules_is_set)


class TestL3AgentNDPProxyExtension(L3AgentNDPProxyTestFramework):

    def _test_router_ndp_proxy(self, enable_ha):
        router_info = self.generate_router_info(enable_ha=enable_ha)
        router_info['enable_ndp_proxy'] = True
        ri = self.manage_router(self.agent, router_info)
        for ndp_proxy in self.ndp_proxies:
            ndp_proxy.router_id = ri.router_id
        (interface_name, namespace,
            iptables_manager) = self.np_ext._get_resource_by_router(ri)
        ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
        self._assert_ndp_proxy_kernel_parameter(ip_wrapper, interface_name)
        self._assert_ndp_iptable_chain_is_set(iptables_manager, interface_name)
        slaac = constants.IPV6_SLAAC
        slaac_mode = {'ra_mode': slaac, 'address_mode': slaac}
        self._add_internal_interface_by_subnet(
            ri.router, count=1, ip_version=constants.IP_VERSION_6,
            ipv6_subnet_modes=[slaac_mode])
        self.agent._process_updated_router(ri.router)
        self._assert_ndp_proxy_state_iptable_rules_is_set(
            ri, iptables_manager, interface_name)
        self._assect_ndp_proxy_rules_is_set(
            ip_wrapper, iptables_manager,
            interface_name, namespace)
        ri.router['enable_ndp_proxy'] = False
        self.agent._process_updated_router(ri.router)

    def test_legacy_router_ndp_proxy(self):
        self._test_router_ndp_proxy(enable_ha=False)

    def test_ha_router_ndp_proxy(self):
        self._test_router_ndp_proxy(enable_ha=True)


class TestL3AgentNDPProxyExtensionDVR(test_dvr_router.TestDvrRouter,
                                      L3AgentNDPProxyTestFramework):

    def test_local_dvr_router(self):
        self.agent.conf.agent_mode = constants.L3_AGENT_MODE_DVR
        router_info = self.generate_dvr_router_info(enable_ha=False)
        for ndp_proxy in self.ndp_proxies:
            ndp_proxy.router_id = router_info['id']
        router_info['enable_ndp_proxy'] = True
        ri = self.manage_router(self.agent, router_info)
        (interface_name, namespace,
            iptables_manager) = self.np_ext._get_resource_by_router(ri)
        ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
        self._assert_ndp_proxy_kernel_parameter(ip_wrapper, interface_name)
        self._assect_ndp_proxy_rules_is_set(ri, interface_name, namespace)

    def test_edge_dvr_router(self):
        self.agent.conf.agent_mode = constants.L3_AGENT_MODE_DVR_SNAT
        router_info = self.generate_dvr_router_info(enable_ha=False)
        for ndp_proxy in self.ndp_proxies:
            ndp_proxy.router_id = router_info['id']
        router_info['enable_ndp_proxy'] = True
        ri = self.manage_router(self.agent, router_info)
        (interface_name, namespace,
            iptables_manager) = self.np_ext._get_resource_by_router(ri)
        ip_wrapper = ip_lib.IPWrapper(namespace=namespace)
        self._assert_ndp_proxy_kernel_parameter(ip_wrapper, interface_name)
        self._assert_ndp_iptable_chain_is_set(iptables_manager, interface_name)
        slaac = constants.IPV6_SLAAC
        slaac_mode = {'ra_mode': slaac, 'address_mode': slaac}
        self._add_internal_interface_by_subnet(
            ri.router, count=1, ip_version=constants.IP_VERSION_6,
            ipv6_subnet_modes=[slaac_mode])
        self.agent._process_updated_router(ri.router)
        self._assert_ndp_proxy_state_iptable_rules_is_set(
            ri, iptables_manager, interface_name)
        super(
            TestL3AgentNDPProxyExtensionDVR,
            self)._assect_ndp_proxy_rules_is_set(
                ip_wrapper, iptables_manager,
                interface_name, namespace)
        ri.router['enable_ndp_proxy'] = False
        self.agent._process_updated_router(ri.router)

    def _assect_ndp_proxy_rules_is_set(self, ri, interface_name, namespace):
        rtr_2_fip_dev = ri.fip_ns.get_rtr_2_fip_device(ri)
        fip_2_rtr_dev = ri.fip_ns.get_fip_2_rtr_device(ri)
        rtr_2_fip_v6_address = self.np_ext._get_device_ipv6_lladdr(
            rtr_2_fip_dev)
        expected_proxy_address = []
        expected_routes = []
        for ndp_proxy in self.ndp_proxies:
            port_binding_obj = self.port_binding_map.get(ndp_proxy.port_id)
            if port_binding_obj and port_binding_obj.host == self.agent.host:
                expected_proxy_address.append(str(ndp_proxy.ip_address))
                expected_routes.append(
                    {'table': 'main', 'source_prefix': None,
                     'cidr': '%s/128' % ndp_proxy.ip_address,
                     'scope': 'global', 'metric': 1024,
                     'proto': 'static', 'device': fip_2_rtr_dev.name,
                     'via': rtr_2_fip_v6_address})

        def check_rules_is_set():
            existing_proxy_addresses = self._get_existing_ndp_proxies(
                interface_name, namespace)
            for address in expected_proxy_address:
                if address not in existing_proxy_addresses:
                    return False

            existing_routes = fip_2_rtr_dev.route.list_routes(
                ip_version=constants.IP_VERSION_6)
            for route in expected_routes:
                if route not in existing_routes:
                    return False

            return True

        common_utils.wait_until_true(check_rules_is_set)



# Copyright 2016 OVH SAS
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.agent.linux import ip_lib
from neutron.common import utils as common_utils
from neutron.tests.fullstack.agents import dhcp_agent
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine
from neutron.tests.unit import testlib_api

load_tests = testlib_api.module_load_tests


class BaseDhcpAgentTest(base.BaseFullStackTestCase):

    scenarios = [
        (constants.AGENT_TYPE_OVS,
         {'l2_agent_type': constants.AGENT_TYPE_OVS}),
    ]
    boot_vm_for_test = True
    dhcp_scheduler_class = None
    api_workers = 1

    def setUp(self):
        host_descriptions = [
            environment.HostDescription(
                dhcp_agent=True,
                l2_agent_type=self.l2_agent_type
            ) for _ in range(self.number_of_hosts)]

        env = environment.Environment(
            environment.EnvironmentDescription(
                l2_pop=False,
                arp_responder=False,
                agent_down_time=self.agent_down_time,
                dhcp_scheduler_class=self.dhcp_scheduler_class,
                api_workers=self.api_workers,
            ),
            host_descriptions)

        super(BaseDhcpAgentTest, self).setUp(env)
        self.project_id = uuidutils.generate_uuid()
        if self.boot_vm_for_test:
            self._create_network_subnet_and_vm()

    def _spawn_vm(self):
        host = random.choice(self.environment.hosts)
        vm = self.useFixture(
            machine.FakeFullstackMachine(
                host,
                self.network['id'],
                self.project_id,
                self.safe_client,
                use_dhcp=True))
        vm.block_until_boot()
        vm.block_until_dhcp_config_done()
        return vm

    def _create_network_subnet_and_vm(self):
        self.network = self.safe_client.create_network(self.project_id)

        self.subnet = self.safe_client.create_subnet(
            self.project_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            name='subnet-test',
            enable_dhcp=True)

        self.vm = self._spawn_vm()


class TestDhcpAgentNoHA(BaseDhcpAgentTest):

    number_of_hosts = 1
    agent_down_time = 60

    def test_mtu_update(self):
        # The test case needs access to devices in nested namespaces. ip_lib
        # doesn't support it, and it's probably unsafe to touch the library for
        # testing matters.
        # TODO(jlibosva) revisit when ip_lib supports nested namespaces
        if self.environment.hosts[0].dhcp_agent.namespace is not None:
            self.skipTest("ip_lib doesn't support nested namespaces")

        self.vm.block_until_dhcp_config_done()

        namespace = dhcp_agent._get_namespace_name(
            self.network['id'],
            suffix=self.environment.hosts[0].dhcp_agent.get_namespace_suffix())
        self.assert_namespace_exists(namespace)

        ip = ip_lib.IPWrapper(namespace)
        devices = ip.get_devices()
        self.assertEqual(1, len(devices))

        dhcp_dev = devices[0]
        mtu = dhcp_dev.link.mtu
        self.assertEqual(1450, mtu)

        mtu -= 1
        self.safe_client.update_network(self.network['id'], mtu=mtu)
        common_utils.wait_until_true(lambda: dhcp_dev.link.mtu == mtu)


class TestDhcpAgentHA(BaseDhcpAgentTest):

    number_of_hosts = 2
    agent_down_time = 30

    def _wait_until_network_rescheduled(self, old_agent):
        def _agent_rescheduled():
            network_agents = self.client.list_dhcp_agent_hosting_networks(
                self.network['id'])['agents']
            if network_agents:
                return network_agents[0]['id'] != old_agent['id']
            return False

        common_utils.wait_until_true(_agent_rescheduled, timeout=120)

    def _kill_dhcp_agent(self, agent):
        for host in self.environment.hosts:
            hostname = host.dhcp_agent.get_agent_hostname()
            if hostname == agent['host']:
                host.dhcp_agent.kill()
                self._wait_until_agent_down(agent['id'])
                break

    def _add_network_to_new_agent(self):
        dhcp_agents = self.client.list_agents(
            agent_type=constants.AGENT_TYPE_DHCP)['agents']
        dhcp_agents_ids = [agent['id'] for agent in dhcp_agents]

        current_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        current_agents_ids = [agent['id'] for agent in current_agents]

        new_agents_ids = list(set(dhcp_agents_ids) - set(current_agents_ids))
        if new_agents_ids:
            new_agent_id = random.choice(new_agents_ids)
            self.client.add_network_to_dhcp_agent(
                new_agent_id, {'network_id': self.network['id']})

    def test_reschedule_network_on_new_agent(self):
        network_dhcp_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        self.assertEqual(1, len(network_dhcp_agents))

        self._kill_dhcp_agent(network_dhcp_agents[0])
        self._wait_until_network_rescheduled(network_dhcp_agents[0])

        # ensure that only one agent is handling DHCP for this network
        new_network_dhcp_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        self.assertEqual(1, len(new_network_dhcp_agents))

        # check if new vm will get IP from new DHCP agent
        self._spawn_vm()

    def test_multiple_agents_for_network(self):
        network_dhcp_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        self.assertEqual(1, len(network_dhcp_agents))

        self._add_network_to_new_agent()
        # ensure that two agents are handling DHCP for this network
        network_dhcp_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        self.assertEqual(2, len(network_dhcp_agents))

        self._kill_dhcp_agent(network_dhcp_agents[0])

        # check if new vm will get IP from DHCP agent which is still alive
        self._spawn_vm()


class TestDhcpAgentHARaceCondition(BaseDhcpAgentTest):

    agent_down_time = 30
    number_of_hosts = 2
    boot_vm_for_test = False
    api_workers = 2
    dhcp_scheduler_class = ('neutron.tests.fullstack.schedulers.dhcp.'
                            'AlwaysTheOtherAgentScheduler')

    def setUp(self):
        super(TestDhcpAgentHARaceCondition, self).setUp()
        self._create_network_with_multiple_subnets()

    def _create_network_with_multiple_subnets(self):
        self.network = self.safe_client.create_network(self.project_id)

        funcs = []
        args = []
        for i in range(4):
            funcs.append(self.safe_client.create_subnet)
            args.append((
                self.project_id, self.network['id'], '10.0.%s.0/24' % i,
                '10.0.%s.1' % i, 'subnet-test-%s' % i, True
            ))
        self._simulate_concurrent_requests_process_and_raise(funcs, args)

    def test_dhcp_agent_ha_with_race_condition(self):
        # NOTE(ralonsoh): the concurrent creation in the same thread could
        # fail because the context and the session is the same for all DB
        # calls.
        network_dhcp_agents = self.client.list_dhcp_agent_hosting_networks(
            self.network['id'])['agents']
        self.assertEqual(1, len(network_dhcp_agents))


class TestSubnetDeleteRace(BaseDhcpAgentTest):
    agent_down_time = 30
    number_of_hosts = 2
    boot_vm_for_test = False

    def setUp(self):
        host_descriptions = [
            environment.HostDescription(
                dhcp_agent=True, l2_agent_type=constants.AGENT_TYPE_OVS),
            environment.HostDescription(
                dhcp_agent=False,
                l2_agent_type=constants.AGENT_TYPE_LINUXBRIDGE)
        ]

        env = environment.Environment(
            environment.EnvironmentDescription(
                network_type='vlan',
                mech_drivers='openvswitch,linuxbridge',
                l2_pop=False,
                arp_responder=False,
                agent_down_time=self.agent_down_time,
                service_plugins='router,segments'
            ),
            host_descriptions)

        # Note(lajoskatona): Here, call grandparent's (BaseFullStackTestCase)
        # setup(), to avoid BaseDhcpAgentTest mess up environment.
        super(BaseDhcpAgentTest, self).setUp(env)
        self.project_id = uuidutils.generate_uuid()
        if self.boot_vm_for_test:
            self._create_network_subnet_and_vm()

    def test_subnet_delete_race_condition(self):
        ovs_physnet = ''
        lb_physnet = ''

        agents = self.client.list_agents()
        for agent in agents['agents']:
            if agent['binary'] == 'neutron-openvswitch-agent':
                ovs_physnet = list(
                    agent['configurations']['bridge_mappings'].keys())[0]
            if agent['binary'] == 'neutron-linuxbridge-agent':
                lb_physnet = list(
                    agent['configurations']['interface_mappings'].keys())[0]

        self.network = self.safe_client.create_network(
            tenant_id=self.project_id, network_type='vlan',
            segmentation_id=103, physical_network=lb_physnet)

        self.segment2 = self.safe_client.create_segment(
            project_id=self.project_id, network=self.network['id'],
            network_type='vlan', name='segment_2', segmentation_id=103,
            physical_network=ovs_physnet)

        subnet = self.safe_client.create_subnet(
            self.project_id,
            self.network['id'],
            cidr='10.0.11.0/24',
            gateway_ip='10.0.11.1',
            name='subnet-test',
            enable_dhcp=True)

        self.vm = self._spawn_vm()

        dhcp_ports = self.safe_client.list_ports(**{
            'device_owner': 'network:dhcp',
            'network_id': self.network['id']
        })
        self.assertEqual(1, len(dhcp_ports))
        self.assertEqual('ACTIVE', dhcp_ports[0]['status'])

        self.vm.destroy(delete_port=True)

        dhcp_ports = self.safe_client.list_ports(**{
            'device_owner': 'network:dhcp',
            'network_id': self.network['id']
        })
        self.assertEqual(1, len(dhcp_ports))

        self.client.delete_subnet(subnet['id'])

        def _is_subnet_deleted():
            snets = self.safe_client.list_subnets()
            if len(snets) == 0:
                return True
            return False

        common_utils.wait_until_true(_is_subnet_deleted)
        # Note(lajoskatona): Here cleanup do its job and the cleanup
        # will fail if the segment or network deletion is inpossible



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_policy import policy

from neutron.conf.policies import base


BINDING_PATH = '/ports/{port_id}/bindings/'
ACTIVATE_BINDING_PATH = '/ports/{port_id}/bindings/{host}'


rules = [
    policy.DocumentedRuleDefault(
        name='get_port_binding',
        check_str=base.ADMIN_OR_SERVICE,
        scope_types=['project'],
        description='Get port binding information',
        operations=[
            {
                'method': 'GET',
                'path': BINDING_PATH,
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='create_port_binding',
        check_str=base.SERVICE,
        scope_types=['project'],
        description='Create port binding on the host',
        operations=[
            {
                'method': 'POST',
                'path': BINDING_PATH,
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='delete_port_binding',
        check_str=base.SERVICE,
        scope_types=['project'],
        description='Delete port binding on the host',
        operations=[
            {
                'method': 'DELETE',
                'path': BINDING_PATH,
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='activate',
        check_str=base.SERVICE,
        scope_types=['project'],
        description='Activate port binding on the host',
        operations=[
            {
                'method': 'PUT',
                'path': ACTIVATE_BINDING_PATH,
            },
        ],
    ),
]


def list_rules():
    return rules



# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

"""delete SecurityGroupPortBinding on security group deletion

Revision ID: 5881373af7f5
Revises: 21ff98fabab1
Create Date: 2022-08-10 07:17:00.360917

"""

# revision identifiers, used by Alembic.
revision = '5881373af7f5'
down_revision = '21ff98fabab1'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.ZED]

TABLE_NAME = 'securitygroupportbindings'


def upgrade():
    inspector = sa.inspect(op.get_bind())
    fk_constraints = inspector.get_foreign_keys(TABLE_NAME)
    for fk in fk_constraints:
        if fk['constrained_columns'] == ['security_group_id']:
            migration.remove_foreign_keys(TABLE_NAME, [fk])
            fk['options']['ondelete'] = 'CASCADE'
            migration.create_foreign_keys(TABLE_NAME, [fk])
            return


def expand_drop_exceptions():
    """Drop the foreign key from "securitygroupportbindings" table

    In order to change the foreign key "security_group_id" from the table
    "securitygroupportbindings" and set the condition "ondelete=CASCADE",
    it is needed first to drop it, modify it and readd it again.
    """
    return {
        sa.ForeignKeyConstraint: [
            'securitygroupportbindings_ibfk_2',  # MySQL name
            'securitygroupportbindings_security_group_id_fkey',  # PGSQL name
        ],
    }



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
import shutil
import tempfile
from unittest import mock

from neutron_lib import fixture as lib_fixtures
from oslo_config import cfg
from oslo_utils import fileutils
from oslo_utils import uuidutils
import psutil

from neutron.agent.linux import external_process as ep
from neutron.common import utils as common_utils
from neutron.tests import base


TEST_UUID = 'test-uuid'
TEST_SERVICE = 'testsvc'
TEST_PID = 1234
TEST_CMDLINE = 'python foo --router_id=%s'
SCRIPT = """#!/bin/bash
output_file=$1
if [ -z "${PROCESS_TAG}" ] ; then
    echo "Variable PROCESS_TAG not set" > $output_file
else
    echo "Variable PROCESS_TAG set: $PROCESS_TAG" > $output_file
fi
"""
DEFAULT_ENVVAR = ep.PROCESS_TAG + '=' + ep.DEFAULT_SERVICE_NAME


class BaseTestProcessMonitor(base.BaseTestCase):

    def setUp(self):
        super(BaseTestProcessMonitor, self).setUp()
        self.log_patch = mock.patch("neutron.agent.linux.external_process."
                                    "LOG.error")
        self.error_log = self.log_patch.start()

        self.spawn_patch = mock.patch("eventlet.spawn")
        self.eventlent_spawn = self.spawn_patch.start()

        # create a default process monitor
        self.create_child_process_monitor('respawn')

    def create_child_process_monitor(self, action):
        conf = mock.Mock()
        conf.AGENT.check_child_processes_action = action
        conf.AGENT.check_child_processes = True
        self.pmonitor = ep.ProcessMonitor(
            config=conf,
            resource_type='test')

    def get_monitored_process(self, uuid, service=None):
        monitored_process = mock.Mock()
        self.pmonitor.register(uuid=uuid,
                               service_name=service,
                               monitored_process=monitored_process)
        return monitored_process


class TestProcessMonitor(BaseTestProcessMonitor):

    def test_error_logged(self):
        pm = self.get_monitored_process(TEST_UUID)
        pm.active = False
        self.pmonitor._check_child_processes()
        self.assertTrue(self.error_log.called)

    def test_exit_handler(self):
        self.create_child_process_monitor('exit')
        pm = self.get_monitored_process(TEST_UUID)
        pm.active = False
        with mock.patch.object(ep.ProcessMonitor,
                               '_exit_handler') as exit_handler:
            self.pmonitor._check_child_processes()
            exit_handler.assert_called_once_with(TEST_UUID, None)

    def test_register(self):
        pm = self.get_monitored_process(TEST_UUID)
        self.assertEqual(len(self.pmonitor._monitored_processes), 1)
        self.assertIn(pm, self.pmonitor._monitored_processes.values())

    def test_register_same_service_twice(self):
        self.get_monitored_process(TEST_UUID)
        self.get_monitored_process(TEST_UUID)
        self.assertEqual(len(self.pmonitor._monitored_processes), 1)

    def test_register_different_service_types(self):
        self.get_monitored_process(TEST_UUID)
        self.get_monitored_process(TEST_UUID, TEST_SERVICE)
        self.assertEqual(len(self.pmonitor._monitored_processes), 2)

    def test_unregister(self):
        self.get_monitored_process(TEST_UUID)
        self.pmonitor.unregister(TEST_UUID, None)
        self.assertEqual(len(self.pmonitor._monitored_processes), 0)

    def test_unregister_unknown_process(self):
        self.pmonitor.unregister(TEST_UUID, None)
        self.assertEqual(len(self.pmonitor._monitored_processes), 0)


class TestProcessManager(base.BaseTestCase):
    def setUp(self):
        super(TestProcessManager, self).setUp()
        self.execute_p = mock.patch('neutron.agent.common.utils.execute')
        self.execute = self.execute_p.start()
        self.delete_if_exists = mock.patch(
            'neutron.agent.linux.utils.delete_if_exists').start()
        self.ensure_dir = mock.patch.object(
            fileutils, 'ensure_tree').start()
        self.error_log = mock.patch("neutron.agent.linux.external_process."
                                    "LOG.error").start()

        self.conf = mock.Mock()
        self.conf.external_pids = '/var/path'

    def test_processmanager_ensures_pid_dir(self):
        pid_file = os.path.join(self.conf.external_pids, 'pid')
        ep.ProcessManager(self.conf, 'uuid', pid_file=pid_file)
        self.ensure_dir.assert_called_once_with(self.conf.external_pids,
                                                mode=0o755)

    def test_enable_no_namespace(self):
        callback = mock.Mock()
        cmd = ['the', 'cmd']
        callback.return_value = cmd

        with mock.patch.object(ep.ProcessManager, 'get_pid_file_name') as name:
            name.return_value = 'pidfile'
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=False)

                manager = ep.ProcessManager(self.conf, 'uuid')
                manager.enable(callback)
                callback.assert_called_once_with('pidfile')
                cmd = ['env', DEFAULT_ENVVAR + '-uuid'] + cmd
                self.execute.assert_called_once_with(cmd,
                                                     check_exit_code=True,
                                                     extra_ok_codes=None,
                                                     run_as_root=False,
                                                     log_fail_as_error=True,
                                                     privsep_exec=False)

    def test_enable_with_namespace(self):
        callback = mock.Mock()
        callback.return_value = ['the', 'cmd']

        with mock.patch.object(ep.ProcessManager, 'get_pid_file_name') as name:
            name.return_value = 'pidfile'
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=False)

                manager = ep.ProcessManager(self.conf, 'uuid', namespace='ns')
                with mock.patch.object(ep, 'ip_lib') as ip_lib:
                    manager.enable(callback)
                    callback.assert_called_once_with('pidfile')
                    self.delete_if_exists.assert_called_once_with(
                        'pidfile', run_as_root=True)
                    env = {ep.PROCESS_TAG: ep.DEFAULT_SERVICE_NAME + '-uuid'}
                    ip_lib.assert_has_calls([
                        mock.call.IPWrapper(namespace='ns'),
                        mock.call.IPWrapper().netns.execute(
                            ['the', 'cmd'], addl_env=env, run_as_root=True)])

    def test_enable_with_namespace_process_active(self):
        callback = mock.Mock()
        callback.return_value = ['the', 'cmd']

        with mock.patch.object(ep.ProcessManager, 'active') as active:
            active.__get__ = mock.Mock(return_value=True)

            manager = ep.ProcessManager(self.conf, 'uuid', namespace='ns')
            with mock.patch.object(ep, 'ip_lib'):
                manager.enable(callback)
                self.assertFalse(callback.called)

    def test_enable_with_ensure_active(self):
        def _create_cmd(*args):
            return ['sleep', 0]

        pm = ep.ProcessManager(self.conf, 'uuid', pid_file='pid_file',
                               default_cmd_callback=_create_cmd)
        with mock.patch.object(psutil, 'Process') as mock_psutil_process, \
                mock.patch.object(ep.ProcessManager, 'pid',
                                  new_callable=mock.PropertyMock) as mock_pid:
            mock_pid.return_value = 'pid_value'
            mock_process = mock.Mock()
            mock_process.cmdline.side_effect = [[], ['the', 'cmd', 'uuid']]
            mock_psutil_process.return_value = mock_process
            try:
                pm.enable(ensure_active=True)
            except common_utils.WaitTimeout:
                self.fail('ProcessManager.enable() raised WaitTimeout')

    def test_enable_delete_pid_file_raises(self):
        callback = mock.Mock()
        cmd = ['the', 'cmd']
        callback.return_value = cmd
        self.delete_if_exists.side_effect = OSError

        with mock.patch.object(ep.ProcessManager, 'get_pid_file_name') as name:
            name.return_value = 'pidfile'
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=False)

                manager = ep.ProcessManager(self.conf, 'uuid')
                manager.enable(callback)
                callback.assert_called_once_with('pidfile')
                cmd = ['env', DEFAULT_ENVVAR + '-uuid'] + cmd
                self.execute.assert_called_once_with(cmd,
                                                     check_exit_code=True,
                                                     extra_ok_codes=None,
                                                     run_as_root=False,
                                                     log_fail_as_error=True,
                                                     privsep_exec=False)
                self.error_log.assert_called_once()

    def test_reload_cfg_without_custom_reload_callback(self):
        with mock.patch.object(ep.ProcessManager, 'disable') as disable:
            manager = ep.ProcessManager(self.conf, 'uuid', namespace='ns')
            manager.reload_cfg()
            disable.assert_called_once_with('HUP', delete_pid_file=False)

    def test_reload_cfg_with_custom_reload_callback(self):
        reload_callback = mock.sentinel.callback
        with mock.patch.object(ep.ProcessManager, 'disable') as disable:
            manager = ep.ProcessManager(
                self.conf, 'uuid', namespace='ns',
                custom_reload_callback=reload_callback)
            manager.reload_cfg()
            disable.assert_called_once_with(get_stop_command=reload_callback,
                                            delete_pid_file=False)

    def test_disable_get_stop_command(self):
        cmd = ['the', 'cmd']
        reload_callback = mock.Mock(return_value=cmd)
        with mock.patch.object(ep.ProcessManager, 'pid',
                               mock.PropertyMock(return_value=4)):
            with mock.patch.object(ep.ProcessManager, 'active',
                                   mock.PropertyMock(return_value=True)):
                manager = ep.ProcessManager(
                    self.conf, 'uuid',
                    custom_reload_callback=reload_callback)
                manager.disable(
                    get_stop_command=manager.custom_reload_callback)
                cmd = ['env', DEFAULT_ENVVAR + '-uuid'] + cmd
                self.assertIn(cmd, self.execute.call_args[0])

    def test_disable_no_namespace(self):
        with mock.patch.object(ep.ProcessManager, 'pid') as pid:
            pid.__get__ = mock.Mock(return_value=4)
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=True)
                manager = ep.ProcessManager(self.conf, 'uuid')

                with mock.patch.object(ep, 'utils') as utils:
                    manager.disable()
                    env = {ep.PROCESS_TAG: ep.DEFAULT_SERVICE_NAME + '-uuid'}
                    utils.assert_has_calls([
                        mock.call.execute(['kill', '-9', 4],
                                          addl_env=env,
                                          run_as_root=False,
                                          privsep_exec=True)])

    def test_disable_namespace(self):
        with mock.patch.object(ep.ProcessManager, 'pid') as pid:
            pid.__get__ = mock.Mock(return_value=4)
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=True)

                manager = ep.ProcessManager(self.conf, 'uuid', namespace='ns')

                with mock.patch.object(ep, 'utils') as utils:
                    manager.disable()
                    env = {ep.PROCESS_TAG: ep.DEFAULT_SERVICE_NAME + '-uuid'}
                    utils.assert_has_calls([
                        mock.call.execute(
                            ['kill', '-9', 4], addl_env=env, run_as_root=True,
                            privsep_exec=True)])

    def test_disable_not_active(self):
        with mock.patch.object(ep.ProcessManager, 'pid') as pid:
            pid.__get__ = mock.Mock(return_value=4)
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=False)
                with mock.patch.object(ep.LOG, 'debug') as debug:
                    manager = ep.ProcessManager(self.conf, 'uuid')
                    manager.disable()
                    debug.assert_called_once_with(mock.ANY, mock.ANY)

    def test_disable_no_pid(self):
        with mock.patch.object(ep.ProcessManager, 'pid') as pid:
            pid.__get__ = mock.Mock(return_value=None)
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=False)
                with mock.patch.object(ep.LOG, 'debug') as debug:
                    manager = ep.ProcessManager(self.conf, 'uuid')
                    manager.disable()
                    debug.assert_called_once_with(mock.ANY, mock.ANY)

    def _test_disable_custom_kill_script(self, kill_script_exists, namespace,
                                         kill_scripts_path='test-path/'):
        cfg.CONF.set_override("kill_scripts_path", kill_scripts_path, "AGENT")
        if kill_script_exists:
            expected_cmd = [
                os.path.join(kill_scripts_path, 'test-service-kill'), '9', 4]
        else:
            expected_cmd = ['kill', '-9', 4]

        with mock.patch.object(ep.ProcessManager, 'pid') as pid:
            pid.__get__ = mock.Mock(return_value=4)
            with mock.patch.object(ep.ProcessManager, 'active') as active:
                active.__get__ = mock.Mock(return_value=True)
                service_name = 'test-service'
                manager = ep.ProcessManager(
                    self.conf, 'uuid', namespace=namespace,
                    service=service_name)
                with mock.patch.object(ep, 'utils') as utils, \
                        mock.patch.object(os.path, 'isfile',
                                          return_value=kill_script_exists):
                    manager.disable()
                    addl_env = {ep.PROCESS_TAG: service_name + '-uuid'}
                    utils.execute.assert_called_with(
                        expected_cmd, addl_env=addl_env,
                        run_as_root=bool(namespace), privsep_exec=True)

    def test_disable_custom_kill_script_no_namespace(self):
        self._test_disable_custom_kill_script(
            kill_script_exists=True, namespace=None)

    def test_disable_custom_kill_script_namespace(self):
        self._test_disable_custom_kill_script(
            kill_script_exists=True, namespace="ns")

    def test_disable_custom_kill_script_no_kill_script_no_namespace(self):
        self._test_disable_custom_kill_script(
            kill_script_exists=False, namespace=None)

    def test_disable_custom_kill_script_no_kill_script_namespace(self):
        self._test_disable_custom_kill_script(
            kill_script_exists=False, namespace="ns")

    def test_disable_custom_kill_script_namespace_no_path(self):
        self._test_disable_custom_kill_script(
            kill_script_exists=False, namespace="ns", kill_scripts_path=None)

    def test_get_pid_file_name_default(self):
        manager = ep.ProcessManager(self.conf, 'uuid')
        retval = manager.get_pid_file_name()
        self.assertEqual(retval, '/var/path/uuid.pid')

    def test_pid(self):
        self.useFixture(lib_fixtures.OpenFixture('/var/path/uuid.pid', '5'))
        manager = ep.ProcessManager(self.conf, 'uuid')
        self.assertEqual(manager.pid, 5)

    def test_pid_no_an_int(self):
        self.useFixture(lib_fixtures.OpenFixture('/var/path/uuid.pid', 'foo'))
        manager = ep.ProcessManager(self.conf, 'uuid')
        self.assertIsNone(manager.pid)

    def test_pid_invalid_file(self):
        with mock.patch.object(ep.ProcessManager, 'get_pid_file_name') as name:
            name.return_value = '.doesnotexist/pid'
            manager = ep.ProcessManager(self.conf, 'uuid')
            self.assertIsNone(manager.pid)

    def test_active(self):
        with mock.patch.object(ep.ProcessManager, 'cmdline') as cmdline:
            cmdline.__get__ = mock.Mock(
                return_value=TEST_CMDLINE % 'uuid')
            manager = ep.ProcessManager(self.conf, 'uuid')
            self.assertTrue(manager.active)

    def test_active_none(self):
        with mock.patch.object(ep.ProcessManager, 'cmdline') as cmdline:
            cmdline.__get__ = mock.Mock(return_value=None)
            manager = ep.ProcessManager(self.conf, 'uuid')
            self.assertFalse(manager.active)

    def test_active_cmd_mismatch(self):
        with mock.patch.object(ep.ProcessManager, 'cmdline') as cmdline:
            cmdline.__get__ = mock.Mock(
                return_value=TEST_CMDLINE % 'anotherid')
            manager = ep.ProcessManager(self.conf, 'uuid')
            self.assertFalse(manager.active)

    def test_cmdline(self):
        with mock.patch.object(psutil, 'Process') as proc:
            proc().cmdline.return_value = (TEST_CMDLINE % 'uuid').split(' ')
            with mock.patch.object(ep.ProcessManager, 'pid') as pid:
                pid.__get__ = mock.Mock(return_value=4)
                manager = ep.ProcessManager(self.conf, 'uuid')
                self.assertEqual(TEST_CMDLINE % 'uuid', manager.cmdline)
        proc().cmdline.assert_called_once_with()

    def test_cmdline_none(self):
        with mock.patch.object(psutil, 'Process') as proc:
            proc.side_effect = psutil.NoSuchProcess(4)
            with mock.patch.object(ep.ProcessManager, 'pid') as pid:
                pid.__get__ = mock.Mock(return_value=4)
                manager = ep.ProcessManager(self.conf, 'uuid')
                self.assertIsNone(manager.cmdline)
        proc.assert_called_once_with(4)


class TestProcessManagerScript(TestProcessManager):
    def setUp(self):
        super().setUp()
        self.env_path = tempfile.mkdtemp(prefix="pm_env_", dir="/tmp/")
        os.chmod(self.env_path, 0o755)
        self.addCleanup(self._clean_env_path)

    def _create_env_var_testing_environment(self, script_content, _create_cmd):
        with tempfile.NamedTemporaryFile('w+', dir=self.env_path,
                                         delete=False) as script:
            script.write(script_content)
        output = tempfile.NamedTemporaryFile('w+', dir=self.env_path,
                                             delete=False)
        os.chmod(script.name, 0o777)
        service_name = 'my_new_service'
        uuid = uuidutils.generate_uuid()
        pm = ep.ProcessManager(self.conf, uuid, service=service_name,
                               default_cmd_callback=_create_cmd)
        return script, output, service_name, uuid, pm

    def _clean_env_path(self):
        shutil.rmtree(self.env_path, ignore_errors=True)

    def test_enable_check_process_id_env_var(self):
        def _create_cmd(*args):
            return [script.name, output.name]

        self.execute_p.stop()
        script, output, service_name, uuid, pm = (
            self._create_env_var_testing_environment(SCRIPT, _create_cmd))
        with mock.patch.object(ep.ProcessManager, 'active') as active:
            active.__get__ = mock.Mock(return_value=False)
            pm.enable()

        with open(output.name, 'r') as f:
            ret_value = f.readline().strip()
        expected_value = ('Variable PROCESS_TAG set: %s-%s' %
                          (service_name, uuid))
        self.assertEqual(expected_value, ret_value)

    def test_disable_check_process_id_env_var(self):
        def _create_cmd(*args):
            return [script.name, output.name]

        self.execute_p.stop()
        script, output, service_name, uuid, pm = (
            self._create_env_var_testing_environment(SCRIPT, _create_cmd))
        with mock.patch.object(ep.ProcessManager, 'active') as active, \
                mock.patch.object(pm, 'get_kill_cmd') as mock_kill_cmd:
            active.__get__ = mock.Mock(return_value=True)
            # NOTE(ralonsoh): the script we are using for testing does not
            # expect to receive the SIG number as the first argument.
            mock_kill_cmd.return_value = [script.name, output.name]
            pm.disable(sig='15')

        with open(output.name, 'r') as f:
            ret_value = f.readline().strip()
        expected_value = ('Variable PROCESS_TAG set: %s-%s' %
                          (service_name, uuid))
        self.assertEqual(expected_value, ret_value)
